import os

import pkg_resources

from kolekto.printer import printer
from kolekto.commands import Command
from kolekto.exceptions import KolektoRuntimeError


DEFAULT_CONFIG = '''
profile = '{profile}'

view 'Titles' {{
    pattern = '{{title}}.{{ext}}'
}}

# By default, the tree will use the tmdb_proxy datasource which will allow
# you to get data from TMDB without requiring an API key:
datasource 'tmdb_proxy' {{
    base_url = 'http://api.kolekto-project.org/'
    max_results = 2
}}

# Uncomment and enter your API key to enable the TMDB datasource:
#datasource 'tmdb' {{
#    api_key = '<enter your tmdb api key>'
#    max_results = 2
#}}

# Get informations from files (quality, runtime...):
datasource 'mediainfos' {{}}

# Uncomment these lines to enable fields rewriting datasource:
#datasource 'rewrite' {{
#    rewrite 'title_upper' {{
#        value = 'movie["title"].upper()'
#    }}
#}}
'''


def get_profiles_name():
    return set(x.name for x in pkg_resources.iter_entry_points(group='kolekto.profiles'))


class Init(Command):

    """ Create links in the kolekto tree.
    """

    help = 'initialize a new Kolekto tree'

    def prepare(self):
        self.add_arg('--profile', '-p', default='movies', choices=get_profiles_name(),
                     help='Select the profile to use')

    def run(self, args, config):
        if config is not None:
            raise KolektoRuntimeError('Already a Kolekto tree')

        movies_directory = os.path.join(args.tree, '.kolekto', 'movies')

        # Ensure that the .kolekto/movies directory exists:
        if not os.path.isdir(movies_directory):
            os.makedirs(movies_directory)

        # Write the default config:
        with open(os.path.join(args.tree, '.kolekto', 'config'), 'w') as fconfig:
            fconfig.write(DEFAULT_CONFIG.format(profile=args.profile))
        printer.p('Initialized empty Kolekto tree in {where}.', where=os.path.abspath(args.tree))

        # Open the metadata db to create it automatically:
        self.get_metadata_db(args.tree)


import json

from kolekto.printer import printer
from kolekto.commands import Command
from kolekto.helpers import get_hash


class Edit(Command):

    """ Edit a movie.
    """

    help = 'edit a movie'

    def prepare(self):
        self.add_arg('input', metavar='movie-hash-or-file')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)

        movie_hash = get_hash(args.input)

        try:
            movie = mdb.get(movie_hash)
        except KeyError:
            printer.p('Unknown movie hash.')
            return

        movie_json = json.dumps(movie, indent=4)

        while True:
            movie_json = printer.edit(movie_json)

            try:
                mdb.save(movie_hash, json.loads(movie_json))
            except ValueError:
                if printer.ask('Bad json data, would you like to try again?', default=True):
                    continue
                else:
                    break
            else:
                printer.p('Saved.')
                break



import unicodedata
import re

from . import Profile


class Movie(dict):

    """ Represent a movie in a dict.
    """

    @property
    def slug(self):
        """ Get a slug from the movie title.
        """
        slug = unicodedata.normalize('NFKD', self['title']).encode('ascii', 'ignore')
        slug = slug.lower().replace(' ', '_')
        slug = re.sub(r'[^a-z0-9_]', '', slug)
        return slug

    def __unicode__(self):
        if self.get('directors'):
            directors = ' by '
            if len(self['directors']) > 1:
                directors += '%s and %s' % (', '.join(self['directors'][0:-1]),
                                                      self['directors'][-1])
            else:
                directors += self['directors'][0]
        else:
            directors = ''
        fmt = u'<b>{title}</b> ({year}){directors}'
        return fmt.format(title=self.get('title', 'No title'),
                          year=self.get('year', 'Unknown'),
                          directors=directors)


class Movies(Profile):

    """ A profile for movies.
    """

    object_class = Movie



import re
import os
import json
from hashlib import sha1
from tempfile import NamedTemporaryFile

from kolekto.printer import printer, option
from kolekto.commands import Command
from kolekto.commands.show import show
from kolekto.datasources import MovieDatasource
from kolekto.exceptions import KolektoRuntimeError


def clean_title(title):
    title = re.sub('[._-]', ' ', title).strip().title()
    match = re.match('(.+)((19|2\d)\d\d)', title)
    if match is None:
        return None, title
    else:
        title = match.group(1)
        year = int(match.group(2))
        return year, title


def copy(tree, source_filename):
    """ Copy file in tree, show a progress bar during operations,
        and return the sha1 sum of copied file.
    """
    #_, ext = os.path.splitext(source_filename)
    filehash = sha1()
    with printer.progress(os.path.getsize(source_filename)) as update:
        with open(source_filename, 'rb') as fsource:
            with NamedTemporaryFile(dir=os.path.join(tree, '.kolekto', 'movies'), delete=False) as fdestination:
                # Copy the source into the temporary destination:
                while True:
                    buf = fsource.read(10 * 1024)
                    if not buf:
                        break
                    filehash.update(buf)
                    fdestination.write(buf)
                    update(len(buf))
                # Rename the file to its final name or raise an error if
                # the file already exists:
                dest = os.path.join(tree, '.kolekto', 'movies', filehash.hexdigest())
                if os.path.exists(dest):
                    raise IOError('This file already exists in tree (%s)' % filehash.hexdigest())
                else:
                    os.rename(fdestination.name, dest)
    return filehash.hexdigest()


def link(tree, source_filename, symlink=False):
    filehash = sha1()
    with printer.progress(os.path.getsize(source_filename)) as update:
        with open(source_filename, 'rb') as fsource:
                # Copy the source into the temporary destination:
                while True:
                    buf = fsource.read(10 * 1024)
                    if not buf:
                        break
                    filehash.update(buf)
                    update(len(buf))
                # Hardlink the file or raise an error if the file already exists:
                dest = os.path.join(tree, '.kolekto', 'movies', filehash.hexdigest())
                if os.path.exists(dest):
                    raise IOError('This file already exists in tree (%s)' % filehash.hexdigest())
                else:
                    if symlink:
                        source_filename = os.path.relpath(source_filename,
                                                          os.path.join(tree, '.kolekto', 'movies'))
                        os.symlink(source_filename, dest)
                    else:
                        os.link(source_filename, dest)
    return filehash.hexdigest()


class BaseImport(Command):

    """ Base class for importers.
    """

    def prepare(self):
        self.add_arg('file', nargs='+', help='Files to import (globbing allowed)')
        self.add_arg('--hardlink', action='store_true', default=False,
                     help='Create an hardlink instead of copying the file')
        self.add_arg('--symlink', action='store_true', default=False,
                     help='Create a symlink instead of copying the file')
        self.add_arg('--auto', '-a', action='store_true', default=False,
                     help='Automatically choose the most revelent movie.')
        self.add_arg('--dont-show', dest='show', action='store_false', default=True,
                     help='Show all informations about imported movie.')
        self.add_arg('--delete', action='store_true', default=False,
                     help='Delete imported file after a successful import')

    def run(self, args, config):
        # Check the args:
        if args.symlink and args.delete:
            raise KolektoRuntimeError('--delete can\'t be used with --symlink')
        elif args.symlink and args.hardlink:
            raise KolektoRuntimeError('--symlink and --hardlink are mutually exclusive')

        # Load the metadata database:
        mdb = self.get_metadata_db(args.tree)

        # Load informations from db:
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)

        for filename in args.file:
            filename = filename.decode('utf8')
            self._import(mdb, mds, args, config, filename)

    def _import(self):
        raise NotImplementedError()


class ImportMovies(BaseImport):

    """ Import movies into the Kolekto tree.
    """

    help = 'import a movie'

    def _import(self, mdb, mds, args, config, filename):
        printer.debug('Importing file {filename}', filename=filename)
        short_filename = os.path.basename(filename)
        title, ext = os.path.splitext(short_filename)

        year, title = clean_title(title)

        # Disable the year filter if auto mode is disabled:
        if not args.auto:
            year = None

        while True:
            # Disable the title input if auto mode is enabled:
            if not args.auto:
                title = printer.input(u'Title to search', default=title)
            datasource, movie = self._search(mds, title, short_filename, year, auto=args.auto)
            if datasource == 'manual':
                movie = self.profile.object_class()
            elif datasource == 'abort':
                printer.p('Aborted import of {filename}', filename=filename)
                return
            break

        if datasource is None:
            return

        # Refresh the full data for the choosen movie:
        movie = mds.refresh(movie)

        if args.show:
            show(movie)
            printer.p('')

        # Edit available data:
        if not args.auto and printer.ask('Do you want to edit the movie metadata', default=False):
            movie = self.profile.object_class(json.loads(printer.edit(json.dumps(movie, indent=True))))

        # Hardlink or copy the movie in the tree
        if args.hardlink or args.symlink:
            printer.p('\nComputing movie sha1sum...')
            movie_hash = link(args.tree, filename, args.symlink)
        else:
            printer.p('\nCopying movie in kolekto tree...')
            movie_hash = copy(args.tree, filename)
        printer.p('')

        mdb.save(movie_hash, movie)
        printer.debug('Movie {hash} saved to the database', hash=movie_hash)

        if args.delete:
            os.unlink(filename)
            printer.debug('Deleted original file {filename}', filename=filename)

    def _search(self, mdb, query, filename, year=None, auto=False):
        """ Search the movie using all available datasources and let the user
            select a result. Return the choosen datasource and produced movie dict.

        If auto is enabled, directly returns the first movie found.
        """
        choices = []
        for datasource, movie in mdb.search(query, year=year):
            if auto:
                return datasource, movie
            if movie.get('directors'):
                directors = ' by '
                if len(movie['directors']) > 1:
                    directors += '%s and %s' % (', '.join(movie['directors'][0:-1]),
                                                          movie['directors'][-1])
                else:
                    directors += movie['directors'][0]
            else:
                directors = ''
            fmt = u'<b>{title}</b> ({year}){directors} [{datasource}]'
            choices.append(option((datasource, movie), fmt, title=movie['title'],
                                                            year=movie.get('year', 'Unknown'),
                                                            directors=directors,
                                                            datasource=datasource.name))

        if not choices:
            printer.p('No results to display for the file: {fn}', fn=filename)
            return None, None

        choices.append(option(('manual', None), 'Enter information manually'))
        choices.append(option(('abort', None), 'None of these'))
        printer.p('Please choose the relevant movie for the file: {fn}', fn=filename, end='\n\n')
        return printer.choice(choices)



import re

from kolekto.datasources import Datasource, DefaultDatasourceSchema

from confiture.schema.containers import Value, Section, many
from confiture.schema.types import String, Boolean


class LazyEval(String):

    """ A string base type evaluating string as Python expression lazily.

    Example in configuration::

        sum = 'sum(range(3, 10))'

    .. warning::
        This type can be dangerous since any Python expression can be typed
        by the user, like __import__("sys").exit(). Use it at your own risk.

    """

    def validate(self, value):
        value = super(LazyEval, self).validate(value)
        def _eval(locals=None, globals=None):
            return eval(value, globals, locals)
        return _eval


class FieldSection(Section):

    _meta = {'repeat': many, 'unique': True, 'args': Value(String())}
    value = Value(LazyEval())
    ignore_error = Value(Boolean(), default=False)


class RewriteDatasourceSchema(DefaultDatasourceSchema):

    rewrite = FieldSection()


class RewriteDatasource(Datasource):

    """ A datasource allowing to rewrite fields.
    """

    config_schema = RewriteDatasourceSchema()

    def __init__(self, *args, **kwargs):
        super(RewriteDatasource, self).__init__(*args, **kwargs)
        self.rewrites = [(x.args, x.get('value'), x.get('ignore_error'))
                         for x in self.config.subsections('rewrite')]

    def attach(self, movie_hash, movie):
        for field, rewrite_func, ignore_error in self.rewrites:
            try:
                value = rewrite_func(globals={'movie': movie, 'sub': re.sub,
                                              'split': re.split})
            except:
                if not ignore_error:
                    raise
            else:
                if value:
                    movie[field] = value
        return movie


import sys
import subprocess
import tempfile

import progressbar


DEFAULT_TAGS = {'b': (lambda **attrs: '\x1b[1m',
                      lambda: '\x1b[21m'),
                'inv': (lambda **attrs: '\x1b[37;7m',
                        lambda :'\x1b[27;39m'),
                'dim': (lambda **attrs: '\x1b[2m',
                        lambda: '\x1b[22m'),
                'u': (lambda **attrs: '\x1b[4m',
                      lambda: '\x1b[24m'),
                '_reset': (None, lambda: '\x1b[0m')}


class ColorHandler(object):

    COLOR_CODE = '\x1b[%dm'
    COLOR_RESET_FG = 39
    COLOR_RESET_BG = 49

    COLORS = {'black': (30, 40),
              'red': (31, 41),
              'green': (32, 42),
              'yellow': (33, 43),
              'blue': (34, 44),
              'magenta': (35, 45),
              'cyan': (36, 46),
              'light_gray': (37, 47),
              'dark_gray': (90, 100),
              'light_red': (91, 101),
              'light_green': (92, 102),
              'light_yellow': (93, 103),
              'light_blue': (94, 104),
              'light_magenta': (95, 105),
              'light_cyan': (96, 106),
              'white': (97, 107)}

    def __init__(self):
        self._last_colors = []

    def open(self, **attrs):
        fg = self.COLORS.get(attrs.get('fg'), [None, None])[0]
        bg = self.COLORS.get(attrs.get('bg'), [None, None])[1]
        self._last_colors.append((fg, bg))
        output = ''
        if fg is not None:
            output += self.COLOR_CODE % fg
        if bg is not None:
            output += self.COLOR_CODE % bg
        return output

    def close(self):
        if self._last_colors:
            self._last_colors.pop()  # Remove colors from our opening tag
            if self._last_colors:
                last_fg, last_bg = self._last_colors[-1]
            else:
                last_fg, last_bg = [None, None]
            if last_fg is None:
                last_fg = self.COLOR_RESET_FG
            if last_bg is None:
                last_bg = self.COLOR_RESET_BG
            output = ''
            output += self.COLOR_CODE % last_fg
            output += self.COLOR_CODE % last_bg
            return output
        else:
            return ''


class ConsoleFormatter(object):

    """ Convert an XML-Like markup to VT100 control codes.
    """

    def __init__(self, tags={}, reset=None):
        self.tags = tags
        self.reset = reset

    def parse(self, text):
        output = ''
        open_tags = []
        for token_type, raw_text, tag_name, attrs in self._lexer(text):
            if token_type == 'text':
                output += raw_text
            elif token_type == 'open':
                open_tags.append(tag_name)
                output += self.tags[tag_name][0](**attrs)
            elif token_type == 'close':
                # Close the actually closed tag, and any other tag opened
                # after and not closed:
                while open_tags:
                    closing_tag = open_tags.pop()
                    output += self.tags[closing_tag][1]()
                    if closing_tag == tag_name:
                        break
        output += self.tags.get('_reset', lambda: '')[1]()
        return output

    def _lexer(self, data):
        while data:
            start = data.find('<')
            if start != -1:
                # Found data before a starting tag
                if start:
                    yield 'text', data[:start], None, None
                    data = data[start:]
                end = data.find('>')
                recheck = data.find('<', 1)
                if recheck > 0 and recheck < end:
                    yield 'text', data[:recheck], None, None
                    data = data[recheck:]
                elif end > 0:
                    tag = data[1:end]
                    # Parse a tag
                    if tag[0] == '/':
                        token_type = 'close'
                        tag = tag[1:]
                    else:
                        token_type = 'open'

                    tag_components = tag.split()
                    if tag_components and tag_components[0] in self.tags:
                        tag_name = tag_components[0]
                        opts = {}
                        for cmpt in tag_components[1:]:
                            if '=' in cmpt:
                                key, value = cmpt.split('=', 1)
                                opts[key] = value
                            else:
                                opts[cmpt] = True
                        yield token_type, data[:end + 1], tag_name, opts
                    else:
                        yield 'text', data[:end + 1], None, None

                    data = data[end + 1:]
                else:
                    break
            else:
                break
        if data:
            yield 'text', data, None, None


class ProgressContext(object):

    """ A context for progress bar allowing to use the with statement.
    """

    def __init__(self, *args, **kwargs):
        self.pbar = progressbar.ProgressBar(*args, **kwargs)
        self.count = 0

    def __call__(self, value):
        self.count += value
        self.pbar.update(self.count)

    def __enter__(self):
        self.pbar.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pbar.finish()


def option(option, text, **kwargs):
    """ Simple shortcut to construct an option for the KolektoPrinter.choices method.
    """
    return option, text, kwargs


class KolektoPrinter(object):

    """ Handle print and user interactions.
    """

    PROGRESS_WIDGETS = (' ', progressbar.Percentage(), ' ',
                        progressbar.Bar(left='[', right=']', marker='='), ' ',
                        progressbar.FileTransferSpeed(), ' | ',
                        progressbar.ETA())
    PROGESS_TASK_WIDGETS = (' ', progressbar.Percentage(), ' ',
                            progressbar.Bar(left='[', right=']', marker='='), ' ',
                            progressbar.SimpleProgress(), ' | ',
                            progressbar.ETA())

    def __init__(self, output=sys.stdout, err=sys.stderr, verbose=False, debug=False, encoding='utf8', editor=None):
        self._output = output
        self._err = err
        self._verbose = verbose
        self._debug = debug
        self._encoding = encoding
        self._editor = editor
        tags = DEFAULT_TAGS.copy()
        color_handler = ColorHandler()
        tags['color'] = (color_handler.open, color_handler.close)
        self.formatter = ConsoleFormatter(tags)

    @property
    def output(self):
        return self._output

    def _print(self, *args, **kwargs):
        sep = kwargs.pop('sep', u' ')
        end = kwargs.pop('end', u'\n')
        err = kwargs.pop('err', False)
        markup = kwargs.pop('markup', True)
        text = sep.join(unicode(x) for x in args).format(**kwargs) + end
        if markup:
            text = self.formatter.parse(text)
        if err:
            self._err.write(text.encode(self._encoding))
        else:
            self._output.write(text.encode(self._encoding))

    def configure(self, output=None, verbose=None, debug=None, encoding=None, editor=None):
        if output is not None:
            self._output = output
        if verbose is not None:
            self._verbose = verbose
        if debug is not None:
            self._debug = debug
        if encoding is not None:
            self._encoding = encoding
        if editor is not None:
            self._editor = editor

    def debug(self, *args, **kwargs):
        if self._debug:
            args = ('[debug]',) + args
            self._print(*args, err=True, **kwargs)

    def verbose(self, *args, **kwargs):
        if self._verbose or self._debug:
            self._print(*args, **kwargs)

    def p(self, *args, **kwargs):
        self._print(*args, **kwargs)

    def choice(self, choices):
        for i, (option, text, args) in enumerate(choices):
            self.p(u'[{indice}] ' + text, indice=i + 1, **args)
        self.p('')
        while True:
            chosen = raw_input('Choice [1-{0}]? '.format(len(choices)))
            if chosen.isdigit() and 1 <= int(chosen) <= len(choices):
                return choices[int(chosen) - 1][0]

    def input(self, prompt, default=None):
        if default is None:
            default_text = u''
        else:
            default_text = u' [{0}]'.format(default)
        prompt = u'{0}{1}? '.format(prompt, default_text)
        answer = raw_input(prompt.encode(self._encoding)).rstrip('\n').decode(self._encoding)
        if not answer:
            answer = default
        return answer

    def ask(self, question, default=False):
        """ Ask a y/n question to the user.
        """
        choices = '[%s/%s]' % ('Y' if default else 'y', 'n' if default else 'N')
        while True:
            response = raw_input('%s %s' % (question, choices)).strip()
            if not response:
                return default
            elif response in 'yYoO':
                return True
            elif response in 'nN':
                return False

    def edit(self, text):
        """ Edit a text using an external editor.
        """
        if isinstance(text, unicode):
            text = text.encode(self._encoding)
        if self._editor is None:
            printer.p('Warning: no editor found, skipping edit')
            return text
        with tempfile.NamedTemporaryFile(mode='w+', suffix='kolekto-edit') as ftmp:
            ftmp.write(text)
            ftmp.flush()
            subprocess.Popen([self._editor, ftmp.name]).wait()
            ftmp.seek(0)
            edited = ftmp.read()
            return edited

    def progress(self, max, task=False):
        if task:
            return ProgressContext(widgets=self.PROGESS_TASK_WIDGETS, maxval=max)
        else:
            return ProgressContext(widgets=self.PROGRESS_WIDGETS, maxval=max)

# Instanciate the default printer:
printer = KolektoPrinter()



import os
import logging

from kolekto.datasources import Datasource
from kolekto.helpers import JsonDbm

import kaa.metadata


# Configure the kaa.metadata logger in order to silent it:
metadata_logger = logging.getLogger('metadata')
metadata_logger.setLevel(logging.CRITICAL)


class MediainfosDatasource(Datasource):

    def __init__(self, *args, **kwargs):
        super(MediainfosDatasource, self).__init__(*args, **kwargs)
        # Open the media infos cache:
        cache_filename = os.path.join(self.tree, '.kolekto', 'media-info-cache.db')
        self._cache = JsonDbm(cache_filename)

    def attach(self, movie_hash, movie):
        if movie_hash in self._cache:
            media_infos = self._cache.get(movie_hash)
        else:
            filename = os.path.join(self.tree, '.kolekto', 'movies', movie_hash)
            infos = kaa.metadata.parse(filename)
            if infos is None:
                return movie

            media_infos = {}
            media_infos['container'] = infos['type'].strip()

            # Set the quality of the video depending on its definition:
            if infos.video[0].width < 1280:
                media_infos['quality'] = 'SD'
            elif 1280 <= infos.video[0].width < 1920:
                media_infos['quality'] = '720p'
            else:
                media_infos['quality'] = '1080p'

            # Set the file extension depending on its mimetype:
            media_infos['ext'] = infos['mime'].split('/')[-1]

            # Set the movie length (in minutes)
            media_infos['runtime'] = int(infos['length'] / 60)
            self._cache.save(movie_hash, media_infos)

        movie.update(media_infos)
        return movie



""" Collection of helpers for Kolekto.
"""

import os
import gdbm
import json

def get_hash(input_string):
    """ Return the hash of the movie depending on the input string.

    If the input string looks like a symbolic link to a movie in a Kolekto
    tree, return its movies hash, else, return the input directly in lowercase.
    """

    # Check if the input looks like a link to a movie:
    if os.path.islink(input_string):
        directory, movie_hash = os.path.split(os.readlink(input_string))
        input_string = movie_hash

    return input_string.lower()


class JsonDbm(object):

    """ A simple GNU DBM database which store JSON-serialized Python.
    """

    def __init__(self, filename, object_class=dict):
        self._db = gdbm.open(filename, 'c')
        self._object_class = object_class

    def __contains__(self, key):
        return key in self._db

    def get(self, key):
        """ Get data associated with provided key.
        """
        return self._object_class(json.loads(self._db[key]))

    def count(self):
        """ Count records in the database.
        """
        return len(self._db)

    def save(self, key, data):
        """ Save data associated with key.
        """
        self._db[key] = json.dumps(data)
        self._db.sync()

    def remove(self, key):
        """ Remove the specified key from the database.
        """
        del self._db[key]
        self._db.sync()

    def iteritems(self):
        """ Iterate over (key, data) couple stored in database.
        """
        for key in self.itermovieshash():
            yield key, self.get(key)


from itertools import chain

import pkg_resources


class Profile(object):

    object_class = dict

    def __init__(self, name, config):
        self.name = name
        self.config = config

    def _get_entrypoints(self):
        return chain(pkg_resources.iter_entry_points(group='kolekto.commands.%s' % self.name),
                     pkg_resources.iter_entry_points(group='kolekto.commands'),
                     pkg_resources.iter_entry_points(group='kolekto.commands.no_profile'))

    def load_commands(self, parser):
        """ Load commands of this profile.

        :param parser: argparse parser on which to add commands
        """

        entrypoints = self._get_entrypoints()

        already_loaded = set()
        for entrypoint in entrypoints:
            if entrypoint.name not in already_loaded:
                command_class = entrypoint.load()
                command_class(entrypoint.name, self, parser).prepare()
                already_loaded.add(entrypoint.name)


class NoProfileProfile(Profile):

    """ Special profile used when no profile is loaded.
    """

    def _get_entrypoints(self):
        return pkg_resources.iter_entry_points(group='kolekto.commands.no_profile')



import json

from kolekto.printer import printer
from kolekto.commands import Command


class Dump(Command):

    """ Dump the whole database into json.
    """

    help = 'dump database into json'

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        dump = [{'hash': x, 'movie': y} for x, y in mdb.itermovies()]
        json.dump(dump, printer.output)


from kolekto.printer import printer
from kolekto.commands import Command
from kolekto.datasources import MovieDatasource
from kolekto.commands.show import show
from kolekto.helpers import get_hash


class Refresh(Command):

    """ Refresh metadata of movies.
    """

    help = 'refresh metadata of movies'

    def prepare(self):
        self.add_arg('input', metavar='movie-hash-or-file', nargs='?',
                     help='Hash or path of the movie to refresh. '
                          'If not specified, refresh all movies.')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)

        if args.input is None: # Refresh all movies
            if printer.ask('Would you like to refresh all movies?', default=True):
                with printer.progress(mdb.count(), task=True) as update:
                    for movie_hash, movie in list(mdb.itermovies()):
                        movie = mds.refresh(movie)
                        mdb.save(movie_hash, movie)
                        printer.verbose('Saved {hash}', hash=movie_hash)
                        update(1)
        else:
            movie_hash = get_hash(args.input)

            try:
                movie = mdb.get(movie_hash)
            except KeyError:
                printer.p('Unknown movie hash.')
                return
            else:
                movie = mds.refresh(movie)
                show(movie)
                if printer.ask('Would you like to save the movie?', default=True):
                    mdb.save(movie_hash, movie)
                    printer.p('Saved.')




import os

from kolekto.printer import printer
from kolekto.commands import Command


class Config(Command):

    """ Edit the configuration file.
    """

    help = 'edit the configuration file'

    def run(self, args, config):
        config_fullname = os.path.join(args.tree, '.kolekto', 'config')
        with open(config_fullname, 'r+') as fconfig:
            config = printer.edit(fconfig.read())
            fconfig.seek(0)
            fconfig.truncate()
            fconfig.write(config)
        printer.p('Saved.')


import os
from itertools import product, izip

from kolekto.printer import printer
from kolekto.commands import Command
from kolekto.datasources import MovieDatasource
from kolekto.pattern import parse_pattern


class FormatWrapper(object):

    """ A wrapper used to customize how movies attributes are formatted.
    """

    def __init__(self, title, obj):
        self.title = title
        self.obj = obj

    def __unicode__(self):
        if isinstance(self.obj, bool):
            if self.obj:
                return self.title.title()
        return unicode(self.obj)


def format_all(format_string, env):
    """ Format the input string using each possible combination of lists
        in the provided environment. Returns a list of formated strings.
    """

    prepared_env = parse_pattern(format_string, env, lambda x, y: [FormatWrapper(x, z) for z in y])
    # Generate each possible combination, format the string with it and yield
    # the resulting string:
    for field_values in product(*prepared_env.itervalues()):
        format_env = dict(izip(prepared_env.iterkeys(), field_values))
        yield format_string.format(**format_env)


def walk_links(directory, prefix='', linkbase=None):
    """ Return all links contained in directory (or any sub directory).
    """
    links = {}
    try:
        for child in os.listdir(directory):
            fullname = os.path.join(directory, child)
            if os.path.islink(fullname):
                link_path = os.path.normpath(os.path.join(directory, os.readlink(fullname)))
                if linkbase:
                    link_path = os.path.relpath(link_path, linkbase)
                links[os.path.join(prefix, child)] = link_path
            elif os.path.isdir(fullname):
                links.update(walk_links(fullname,
                                        prefix=os.path.join(prefix, child),
                                        linkbase=linkbase))
    except OSError as err:
        if err.errno != 2:  # Ignore unknown directory error
            raise
    return links


class Link(Command):

    """ Create links in the kolekto tree.
    """

    help = 'create symlinks'

    def prepare(self):
        self.add_arg('--dry-run', '-d', action='store_true', default=False,
                     help='Do not create or delete any link')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)

        if args.dry_run:
            printer.p('Dry run: I will not create or delete any link')

        # Create the list of links that must exists on the fs:
        db_links = {}
        with printer.progress(mdb.count(), task=True) as update:
            for movie_hash, movie in mdb.itermovies():
                movie = mds.attach(movie_hash, movie)
                for view in config.subsections('view'):
                    for pattern in view.get('pattern'):
                        for result in format_all(pattern, movie):
                            filename = os.path.join(view.args, result)
                            if filename in db_links:
                                printer.p('Warning: duplicate link {link}', link=filename)
                            else:
                                db_links[filename] = movie_hash
                update(1)

        # Create the list of links already existing on the fs:
        fs_links = {}
        for view in config.subsections('view'):
            view_links = walk_links(os.path.join(args.tree, view.args),
                                    prefix=view.args,
                                    linkbase=os.path.join(args.tree, '.kolekto', 'movies'))
            fs_links.update(view_links)

        db_links = set(db_links.iteritems())
        fs_links = set(fs_links.iteritems())

        links_to_delete = fs_links - db_links
        links_to_create = db_links - fs_links

        printer.p('Found {rem} links to delete, {add} links to create',
                  rem=len(links_to_delete),
                  add=len(links_to_create))

        dirs_to_cleanup = set()

        # Delete the old links:
        for filename, link in links_to_delete:
            printer.verbose('Deleting {file}', file=filename)
            if not args.dry_run:
                os.remove(os.path.join(args.tree, filename))
            while filename:
                filename = os.path.split(filename)[0]
                dirs_to_cleanup.add(filename)

        dirs_to_cleanup.discard('')  # Avoid to delete view roots

        # Delete empty directories:
        for directory in dirs_to_cleanup:
            if not args.dry_run:
                try:
                    os.rmdir(os.path.join(args.tree, directory))
                except OSError, err:
                    if err.errno != 39:  # Ignore "Directory not empty" error
                        raise
                else:
                    printer.verbose('Deleted directory {dir}', dir=directory)

        # Create the new links:
        for filename, link in links_to_create:
            fullname = os.path.join(args.tree, filename)
            dirname = os.path.dirname(fullname)
            movie_link = os.path.join(args.tree, '.kolekto', 'movies', link)
            printer.verbose('Link: {file!r} -> {link!r}', file=filename, link=link)
            if not args.dry_run:
                try:
                    os.makedirs(dirname)
                except OSError as err:
                    if err.errno != 17:  # Ignore already existing directory error
                        raise
                os.symlink(os.path.relpath(movie_link, dirname), fullname)


from kolekto.commands import Command
from kolekto.printer import printer
from kolekto.helpers import get_hash


class FlagCommand(Command):

    flags = []  # Flags to set when this flag is set
    unset_flags = []  # Flags to unset when this flag is set
    unflag_unset_flags = []  # Flags to unset when this flag is unset

    def prepare(self):
        self.add_arg('input', metavar='movie-hash-or-file')
        self.add_arg('--unflag', '-u', action='store_true', default=False)

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)

        movie_hash = get_hash(args.input)

        try:
            movie = mdb.get(movie_hash)
        except KeyError:
            printer.p('Unknown movie hash.')
            return
        if args.unflag:
            for flag in self.unflag_unset_flags:
                try:
                    del movie[flag]
                except KeyError:
                    pass
        else:
            for flag in self.flags:
                movie[flag] = True
            for flag in self.unset_flags:
                try:
                    del movie[flag]
                except KeyError:
                    pass
        mdb.save(movie_hash, movie)


class Watch(FlagCommand):

    """ Flag a movie as watched.
    """

    flags = ['watched']
    unflag_unset_flags = ['watched', 'favorite', 'crap']
    help = 'flag a movie as watched'


class Favorite(FlagCommand):

    """ Flag a movie as favorite.
    """

    flags = ['favorite', 'watched']
    unset_flags = ['crap']
    unflag_unset_flags = ['favorite']
    help = 'flag a movie as favorite (and also set it as watched)'


class Crap(FlagCommand):

    """ Flag a movie as crap.
    """

    flags = ['crap', 'watched']
    unset_flags = ['favorite']
    unflag_unset_flags = ['crap']
    help = 'flag a movie as crap (and also set it as watched)'



from kolekto.commands import Command
from kolekto.printer import printer
from kolekto.exceptions import KolektoRuntimeError
from kolekto.pattern import parse_pattern
from kolekto.datasources import MovieDatasource


DEFAULT_PATTERN = u'<b>{title}</b> ({year|"unknown"}) by {directors}'
DEFAULT_ORDER = ('title', 'year')


class ListingFormatWrapper(object):

    """ A wrapper used to customize how movies attributes are formatted.
    """

    def __init__(self, title, obj):
        self.title = title
        self.obj = obj

    def __unicode__(self):
        if isinstance(self.obj, bool):
            if self.obj:
                return self.title.title()
        elif isinstance(self.obj, list):
            if not self.obj:
                return 'None'  # List is empty
            elif len(self.obj) > 1:
                return '%s and %s' % (', '.join(self.obj[0:-1]), self.obj[-1])
            else:
                return unicode(self.obj[0])
        else:
            return unicode(self.obj)

    def __repr__(self):
        return repr(self.obj)


class List(Command):

    """ List movies in the kolekto tree.
    """

    help = 'list movies'

    def prepare(self):
        self.add_arg('listing', metavar='listing', default='default', nargs='?')

    def _config(self, args, config):
        """ Get configuration for the current used listing.
        """
        listings = dict((x.args, x) for x in config.subsections('listing'))
        listing = listings.get(args.listing)
        if listing is None:
            if args.listing == u'default':
                return {'pattern': DEFAULT_PATTERN, 'order': DEFAULT_ORDER}
            else:
                raise KolektoRuntimeError('Unknown listing %r' % args.listing)
        else:
            return {'pattern': listing.get('pattern'),
                    'order': listing.get('order')}

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)
        listing = self._config(args, config)
        def _sorter((movie_hash, movie)):
            return tuple(movie.get(x) for x in listing['order'])
        movies = sorted(mdb.itermovies(), key=_sorter)
        # Get the current used listing:
        for movie_hash, movie in movies:
            movie = mds.attach(movie_hash, movie)
            prepared_env = parse_pattern(listing['pattern'], movie, ListingFormatWrapper)
            printer.p(u'<inv><b> {hash} </b></inv> ' + listing['pattern'], hash=movie_hash, **prepared_env)


""" This module provide a small TMDB proxy.
"""

import os
import json
from datetime import datetime

import requests
import redis

from flask import Flask, Response, request


TMDB_API_URL = u'http://api.themoviedb.org/3'
ONE_WEEK = 604800


app = Flask(__name__)
app.config.update(REDIS_HOST=os.environ.get('KOLEKTO_REDIS_HOST', 'localhost'),
                  REDIS_PORT=int(os.environ.get('KOLEKTO_REDIS_PORT', 6379)),
                  REDIS_DB=int(os.environ.get('KOLEKTO_REDIS_DB', 0)),
                  REDIS_RO_HOST=os.environ.get('KOLEKTO_REDIS_RO_HOST', None),
                  REDIS_RO_PORT=int(os.environ.get('KOLEKTO_REDIS_RO_PORT', 6379)),
                  TMDB_API_KEY=os.environ['KOLEKTO_TMDB_API_KEY'],
                  CACHE_TTL=int(os.environ.get('KOLEKTO_CACHE_TTL', ONE_WEEK)),
                  DEBUG=os.environ.get('KOLEKTO_DEBUG') == 'on')

redis_conn = redis.StrictRedis(host=app.config['REDIS_HOST'],
                               port=app.config['REDIS_PORT'],
                               db=app.config['REDIS_DB'])
if app.config['REDIS_RO_HOST'] is not None:
    redis_ro_conn = redis.StrictRedis(host=app.config['REDIS_RO_HOST'],
                                      port=app.config['REDIS_RO_PORT'],
                                      db=app.config['REDIS_DB'])
else:
    redis_ro_conn = redis_conn
requests_session = requests.Session()


def get_on_tmdb(uri, **kwargs):
    """ Get a resource on TMDB.
    """
    kwargs['api_key'] = app.config['TMDB_API_KEY']
    response = requests_session.get((TMDB_API_URL + uri).encode('utf8'), params=kwargs)
    response.raise_for_status()
    return json.loads(response.text)


@app.route('/1/search')
def search():
    """ Search a movie on TMDB.
    """
    redis_key = 's_%s' % request.args['query'].lower()
    cached = redis_ro_conn.get(redis_key)
    if cached:
        return Response(cached)
    else:
        try:
            found = get_on_tmdb(u'/search/movie', query=request.args['query'])
            movies = []
            for movie in found['results']:
                cast = get_on_tmdb(u'/movie/%s/casts' % movie['id'])
                year = datetime.strptime(movie['release_date'], '%Y-%m-%d').year if movie['release_date'] else None
                movies.append({'title': movie['original_title'],
                               'directors': [x['name'] for x in cast['crew'] if x['department'] == 'Directing'],
                               'year': year,
                               '_tmdb_id': movie['id']})
        except requests.HTTPError as err:
            return Response('TMDB API error: %s' % str(err), status=err.response.status_code)
        json_response = json.dumps({'movies': movies})
        redis_conn.setex(redis_key, app.config['CACHE_TTL'], json_response)
        return Response(json_response)


@app.route('/1/movie/<int:tmdb_id>')
def get_movie(tmdb_id):
    """ Get informations about a movie using its tmdb id.
    """
    redis_key = 'm_%s' % tmdb_id
    cached = redis_ro_conn.get(redis_key)
    if cached:
        return Response(cached)
    else:
        try:
            details = get_on_tmdb(u'/movie/%d' % tmdb_id)
            cast = get_on_tmdb(u'/movie/%d/casts' % tmdb_id)
            alternative = get_on_tmdb(u'/movie/%d/alternative_titles' % tmdb_id)
        except requests.HTTPError as err:
            return Response('TMDB API error: %s' % str(err), status=err.response.status_code)
        movie = {'title': details['original_title'],
                 'score': details['popularity'],
                 'directors': [x['name'] for x in cast['crew'] if x['department'] == 'Directing'],
                 'writers': [x['name'] for x in cast['crew'] if x['department'] == 'Writing'],
                 'cast': [x['name'] for x in cast['cast']],
                 'genres': [x['name'] for x in details['genres']],
                 'countries': [x['name'] for x in details['production_countries']],
                 'tmdb_votes': int(round(details.get('vote_average', 0) * 0.5)),
                 '_tmdb_id': tmdb_id}
        if details.get('release_date'):
            movie['year'] = datetime.strptime(details['release_date'], '%Y-%m-%d').year
        if details.get('belongs_to_collection'):
            movie['collection'] = details['belongs_to_collection']['name']
        for alt in alternative['titles']:
            movie['title_%s' % alt['iso_3166_1'].lower()] = alt['title']
        json_response = json.dumps({'movie': movie})
        redis_conn.setex(redis_key, app.config['CACHE_TTL'], json_response)
        return Response(json_response)


if __name__ == '__main__':
    app.run()


import os

from kolekto.printer import printer
from kolekto.commands import Command


class Gc(Command):

    """ Garbage collect orphan files stored in tree.
    """

    help = 'garbage collect orphan files'

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        db_files = set()
        for movie_hash, movie in mdb.itermovies():
            db_files.add(movie_hash)
            db_files.update(movie.get('_externals', []))
        printer.verbose('Found {nb} files in database', nb=len(db_files))
        fs_files = set(os.listdir(os.path.join(args.tree, '.kolekto', 'movies')))
        printer.verbose('Found {nb} files in filesystem', nb=len(fs_files))
        orphan_files = fs_files - db_files
        printer.p('Found {nb} orphan files to delete', nb=len(orphan_files))
        if orphan_files:
            printer.verbose('Files to delete: {files}', files=', '.join(orphan_files))
            if printer.ask('Would you like to delete orphans?'):
                for orphan_file in orphan_files:
                    try:
                        os.remove(os.path.join(args.tree, '.kolekto', 'movies', orphan_file))
                    except OSError as err:
                        printer.p('Unable to delete {file}: {err}', file=orphan_file, err=err)
                    else:
                        printer.verbose('Deleted {file}', file=orphan_file)


""" The moviedb database for Kolekto.
"""

import json
import requests
import time
from datetime import datetime

from kolekto.printer import printer
from kolekto.datasources import Datasource, DefaultDatasourceSchema
from kolekto.exceptions import KolektoRuntimeError

from confiture.schema.containers import Value
from confiture.schema.types import String, Integer


requests_session = requests.Session()


class TmdbDatasourceSchema(DefaultDatasourceSchema):

    api_key = Value(String())
    max_results = Value(Integer(min=1), default=None)


class TmdbDatasource(Datasource):

    config_schema = TmdbDatasourceSchema()

    URL_SEARCH = (u'http://api.themoviedb.org/3/search/movie'
                   '?api_key=%(api_key)s&query=%(query)s')
    URL_GET = u'http://api.themoviedb.org/3/movie/%(id)s?api_key=%(api_key)s'
    URL_CAST = u'http://api.themoviedb.org/3/movie/%(id)s/casts?api_key=%(api_key)s'
    URL_ALT = u'http://api.themoviedb.org/3/movie/%(id)s/alternative_titles?api_key=%(api_key)s'

    def _tmdb_search(self, title):
        ak = self.config.get('api_key')
        url = self.URL_SEARCH % dict(api_key=ak, query=requests.utils.quote(title.encode('utf8')))
        response = requests_session.get(url.encode('utf8'))
        return json.loads(response.text)

    def _get(self, url, *args, **kwargs):
        ak = self.config.get('api_key')
        url = url % dict(api_key=ak, **kwargs)
        for _ in xrange(3):
            printer.debug('Requesting {url}', url=url)
            response = requests.get(url)
            if not response.ok:
                printer.debug('Got error ({http_err}), retrying in 3s...', http_err=response.status_code)
                time.sleep(3)
                continue
            return json.loads(response.text)
        else:
            err_msg = 'Unable to get the URL, server reported error: %s' % response.status_code
            # Show the error reason message to the user if the requests
            # version is sufficiently recent:
            if hasattr(response, 'reason'):
                err_msg += ' (%s)' % response.reason
            raise KolektoRuntimeError(err_msg)

    def _tmdb_get(self, movie_id):
        return self._get(self.URL_GET, id=movie_id)

    def _tmdb_cast(self, movie_id):
        return self._get(self.URL_CAST, id=movie_id)

    def _tmdb_alt(self, movie_id):
        return self._get(self.URL_ALT, id=movie_id)

    def search(self, title, year=None):
        results = self._tmdb_search(title)['results']
        max_results = self.config.get('max_results')
        for result in results:
            # Abort the search if max results is reached:
            if max_results == 0:
                break

            # Parse the release year:
            if result.get('release_date'):
                movie_year = datetime.strptime(result['release_date'], '%Y-%m-%d').year
            else:
                movie_year = None

            # Skip the movie if searched date is not the release date:
            if year is not None and year != movie_year:
                continue

            # Else, format and yield the movie:
            cast = self._tmdb_cast(result['id'])
            movie = self.object_class({'title': result['original_title'],
                                       'directors': [x['name'] for x in cast['crew'] if x['department'] == 'Directing'],
                                       '_datasource': self.name,
                                       '_tmdb_id': result['id']})
            if movie_year:
                movie['year'] = movie_year
            yield movie

            if max_results is not None:
                max_results -= 1

    def refresh(self, movie):
        """ Try to refresh metadata of the movie through the datasource.
        """
        if '_tmdb_id' in movie:
            tmdb_id = movie['_tmdb_id']
            details = self._tmdb_get(tmdb_id)
            cast = self._tmdb_cast(tmdb_id)
            alternatives = self._tmdb_alt(tmdb_id)
            refreshed = self.object_class({'title': details['original_title'],
                                           'score': details['popularity'],
                                           'directors': [x['name'] for x in cast['crew'] if x['department'] == 'Directing'],
                                           'writers': [x['name'] for x in cast['crew'] if x['department'] == 'Writing'],
                                           'cast': [x['name'] for x in cast['cast']],
                                           'genres': [x['name'] for x in details['genres']],
                                           'countries': [x['name'] for x in details['production_countries']],
                                           'tmdb_votes': int(round(details.get('vote_average', 0) * 0.5)),
                                           '_Datasource': self.name,
                                           '_tmdb_id': tmdb_id})
            if details.get('release_date'):
                refreshed['year'] = datetime.strptime(details['release_date'], '%Y-%m-%d').year
            if details.get('belongs_to_collection'):
                refreshed['collection'] = details['belongs_to_collection']['name']
            for alt in alternatives['titles']:
                refreshed['title_%s' % alt['iso_3166_1'].lower()] = alt['title']
            return refreshed


class TmdbProxyDatasourceSchema(DefaultDatasourceSchema):

    base_url = Value(String())
    max_results = Value(Integer(min=1), default=None)


class TmdbProxyDatasource(Datasource):

    config_schema = TmdbProxyDatasourceSchema()

    def _get(self, uri, *args, **kwargs):
        url = self.config.get('base_url').rstrip('/') + uri
        for _ in xrange(3):
            printer.debug('Requesting {url}', url=url)
            response = requests_session.get(url, params=kwargs)
            if not response.ok :
                printer.debug('Got error ({http_err}), retrying in 3s...', http_err=response.status_code)
                time.sleep(3)
                continue
            return json.loads(response.text)
        else:
            err_msg = 'Unable to get the URL, server reported error: %s' % response.status_code
            # Show the error reason message to the user if the requests
            # version is sufficiently recent:
            if hasattr(response, 'reason'):
                err_msg += ' (%s)' % response.reason
            raise KolektoRuntimeError(err_msg)

    def search(self, title, year=None):
        results = self._get('/1/search', query=title)['movies']
        max_results = self.config.get('max_results')

        for result in results:
            # Abort the search if max results is reached:
            if max_results == 0:
                break

            # Skip the movie if searched date is not the release date:
            if year is not None and year != result.get('year'):
                continue

            movie = self.object_class(result)
            yield movie

            if max_results is not None:
                max_results -= 1

    def refresh(self, movie):
        """ Try to refresh metadata of the movie through the datasource.
        """
        if '_tmdb_id' in movie:
            tmdb_id = movie['_tmdb_id']
            movie = self._get('/1/movie/%s' % tmdb_id)
            return movie['movie']



from kolekto.helpers import JsonDbm


class MoviesMetadata(JsonDbm):

    """ A database used to store metadata about movies managed by kolekto.
    """

    def itermovieshash(self):
        """ Iterate over movies hash stored in the database.
        """
        cur = self._db.firstkey()
        while cur is not None:
            yield cur
            cur = self._db.nextkey(cur)

    def itermovies(self):
        """ Iterate over (hash, movie) couple stored in database.
        """
        return self.iteritems()


from collections import defaultdict

from kolekto.printer import printer
from kolekto.commands import Command
from kolekto.datasources import MovieDatasource


class FindDuplicates(Command):

    """ Find duplicate movies.
    """

    help = 'find duplicate movies in collection'

    def run(self, args, config):
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)
        mdb = self.get_metadata_db(args.tree)
        hash_by_title = defaultdict(lambda: [])
        for movie_hash, movie in mdb.itermovies():
            movie = mds.attach(movie_hash, movie)
            hash_info = '<inv> %s </inv> (%s/%s)' % (movie_hash, movie.get('quality'), movie.get('ext'))
            hash_by_title[movie.get('title', None), movie.get('year', None)].append(hash_info)

        for (title, year), hashs in hash_by_title.iteritems():
            if len(hashs) > 1:
                printer.p('<b>{title}</b> ({year}): {hashs}', title=title,
                          year=year, hashs=' '.join(hashs))


""" Parse and format patterns.
"""

from string import Formatter


def parse_pattern(format_string, env, wrapper=lambda x, y: y):
    """ Parse the format_string and return prepared data according to the env.

    Pick each field found in the format_string from the env(ironment), apply
    the wrapper on each data and return a mapping between field-to-replace and
    values for each.
    """

    formatter = Formatter()
    fields = [x[1] for x in formatter.parse(format_string) if x[1] is not None]

    prepared_env = {}

    # Create a prepared environment with only used fields, all as list:
    for field in fields:
        # Search for a movie attribute for each alternative field separated
        # by a pipe sign:
        for field_alt in (x.strip() for x in field.split('|')):
            # Handle default values (enclosed by quotes):
            if field_alt[0] in '\'"' and field_alt[-1] in '\'"':
                field_values = field_alt[1:-1]
            else:
                field_values = env.get(field_alt)
            if field_values is not None:
                break
        else:
            field_values = []
        if not isinstance(field_values, list):
            field_values = [field_values]
        prepared_env[field] = wrapper(field_alt, field_values)

    return prepared_env



from kolekto.commands import Command
from kolekto.printer import printer
from kolekto.helpers import get_hash


class Rm(Command):

    """ Remove a movie from the kolekto tree.
    """

    help = 'remove a movie'

    def prepare(self):
        self.add_arg('input', metavar='movie-hash-or-file')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)

        movie_hash = get_hash(args.input)

        try:
            mdb.get(movie_hash)
        except KeyError:
            printer.p('Unknown movie hash.')
            return

        if printer.ask('Are you sure?', default=False):
            mdb.remove(movie_hash)
            printer.p('Removed. You need to launch gc to free space.')


""" Kolekto configuration parsers.
"""

import pkg_resources

from confiture import Confiture
from confiture.schema import ValidationError
from confiture.schema.containers import Section, Value, List
from confiture.schema.types import String

from .profiles.movies import Movies


class Profile(String):

    """ Type for Kolekto profiles.

    This type load know profiles using distutils entry-points.
    """

    def validate(self, value):
        value = super(Profile, self).validate(value)
        # Get the first profile to match the provided value:
        profile = next(pkg_resources.iter_entry_points('kolekto.profiles', str(value)), None)
        if profile is None:
            raise ValidationError('Unknown profile')
        return profile.name, profile.load()

    def cast(self, value):
        raise NotImplementedError()


class ViewKolektoConfig(Section):

    _meta = {'args': Value(String()),
             'unique': True,
             'repeat': (0, None)}

    pattern = List(String())


class DatasourceKolektoConfig(Section):

    _meta = {'args': Value(String()),
             'unique': True,
             'repeat': (0, None),
             'allow_unknown': True}


class ListingKolektoConfig(Section):

    _meta = {'args': Value(String()),
             'unique': True,
             'repeat': (0, None)}

    pattern = Value(String())
    order = List(String(), default=['title', 'year'])


class RootKolektoConfig(Section):

    profile = Value(Profile(), default=('movies', Movies))
    view = ViewKolektoConfig()
    datasource = DatasourceKolektoConfig()
    listing = ListingKolektoConfig()


def parse_config(filename):
    conf = Confiture.from_filename(filename, schema=RootKolektoConfig())
    return conf.parse()



from kolekto.commands import Command
from kolekto.datasources import MovieDatasource
from kolekto.printer import printer
from kolekto.helpers import get_hash


METADATA_SORTER_FIRST = ('title', 'year', 'directors', 'cast', 'writers',
                         'collection', 'genres')


def show(movie):
    """ Show the movie metadata.
    """
    for key, value in sorted(movie.iteritems(), cmp=metadata_sorter, key=lambda x: x[0]):
        if isinstance(value, list):
            if not value:
                continue
            other = value[1:]
            value = value[0]
        else:
            other = []
        printer.p('<b>{key}</b>: {value}', key=key, value=value)
        for value in other:
            printer.p('{pad}{value}', value=value, pad=' ' * (len(key) + 2))


def metadata_sorter(x, y):
    """ Sort metadata keys by priority.
    """
    if x == y:
        return 0
    if x in METADATA_SORTER_FIRST and y in METADATA_SORTER_FIRST:
        return -1 if METADATA_SORTER_FIRST.index(x) < METADATA_SORTER_FIRST.index(y) else 1
    elif x in METADATA_SORTER_FIRST:
        return -1
    elif y in METADATA_SORTER_FIRST:
        return 1
    else:
        if x.startswith('_') and y.startswith('_'):
            return cmp(x[1:], y[1:])
        elif x.startswith('_'):
            return 1
        elif y.startswith('_'):
            return -1
        else:
            return cmp(x, y)


class Show(Command):

    """ Show information about movies.
    """

    help = 'show informations about a movie'

    def prepare(self):
        self.add_arg('input', metavar='movie-hash-or-file')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)

        movie_hash = get_hash(args.input)

        try:
            movie = mdb.get(movie_hash)
        except KeyError:
            printer.p('Unknown movie hash.')
            return
        movie = mds.attach(movie_hash, movie)
        show(movie)


import json

from kolekto.printer import printer
from kolekto.commands import Command


class Restore(Command):

    """ Restore metadata from a json dump.
    """

    help = 'Restore metadata from a json dump'

    def prepare(self):
        self.add_arg('file', help='The json dump file to restore')

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        with open(args.file) as fdump:
            dump = json.load(fdump)
        for movie in dump:
            mdb.save(movie['hash'], movie['movie'])
            printer.verbose('Loaded {hash}', hash=movie['hash'])
        printer.p('Loaded {nb} movies.', nb=len(dump))



import os
from datetime import timedelta
from collections import defaultdict
from itertools import islice

from kolekto.commands import Command
from kolekto.datasources import MovieDatasource
from kolekto.printer import printer


SUFFIXES = ('KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')


def humanize_filesize(value):
    """ Return an humanized file size.
    """

    value = float(value)

    if value == 1:
        return '1 Byte'
    elif value < 1024:
        return '%d Bytes' % value
    elif value < 1024:
        return '%dB' % value

    for i, s in enumerate(SUFFIXES):
        unit = 1024 ** (i + 2)
        if value < unit:
            return '%.1f %s' % ((1024 * value / unit), s)
    return '%.1f %s' % ((1024 * value / unit), s)


def format_top(counter, top=3):
    """ Format a top.
    """
    items = islice(reversed(sorted(counter.iteritems(), key=lambda x: x[1])), 0, top)
    return u'; '.join(u'{g} ({nb})'.format(g=g, nb=nb) for g, nb in items)


class Stats(Command):

    """ Get stats about the movies collection.
    """

    help = 'get stats about the movie collection'

    def run(self, args, config):
        mdb = self.get_metadata_db(args.tree)
        mds = MovieDatasource(config.subsections('datasource'), args.tree, self.profile.object_class)
        total_runtime = 0
        total_size = 0
        count_by_genre = defaultdict(lambda: 0)
        count_by_director = defaultdict(lambda: 0)
        count_by_quality = defaultdict(lambda: 0)
        count_by_container = defaultdict(lambda: 0)
        for movie_hash, movie in mdb.itermovies():
            movie_fullpath = os.path.join(args.tree, '.kolekto', 'movies', movie_hash)
            movie = mds.attach(movie_hash, movie)
            total_runtime += movie.get('runtime', 0)
            total_size += os.path.getsize(movie_fullpath)
            for genre in movie.get('genres', []):
                count_by_genre[genre] += 1
            for director in movie.get('directors', []):
                count_by_director[director] += 1
            count_by_quality[movie.get('quality', 'n/a')] += 1
            count_by_container[movie.get('container', 'n/a')] += 1

        printer.p('<b>Number of movies:</b>', mdb.count())
        printer.p('<b>Total runtime:</b>', timedelta(seconds=total_runtime * 60))
        printer.p('<b>Total size:</b>', humanize_filesize(total_size))
        printer.p('<b>Genres top3:</b>', format_top(count_by_genre))
        printer.p('<b>Director top3:</b>', format_top(count_by_director))
        printer.p('<b>Quality:</b>', format_top(count_by_quality, None))
        printer.p('<b>Container:</b>', format_top(count_by_container, None))


""" Movies database datasources.
"""

import pkg_resources

from confiture.schema.containers import Section, Value
from confiture.schema.types import String

from ..exceptions import KolektoRuntimeError


class DefaultDatasourceSchema(Section):

    _meta = {'args': Value(String()),
             'unique': True,
             'repeat': (0, None),
             'allow_unknown': True}


class Datasource(object):

    """ Base class for all movies database datasources.
    """

    config_schema = DefaultDatasourceSchema()

    def __init__(self, name, tree, config, object_class=dict):
        self.name = name
        self.tree = tree
        self.config = config
        self.object_class = object_class

    def search(self, title, year=None):
        """ Search for a movie title in database.

        :return: a list of Movie object
        """
        return []

    def refresh(self, movie):
        """ Refresh the movie's metadata.
        """

    def attach(self, movie_hash, movie):
        return movie


class MovieDatasource(object):

    """ Movie database.
    """

    def __init__(self, datasources_config, tree, object_class=dict):
        self._object_class = object_class
        self._datasources = []
        for datasource_config in datasources_config:
            entrypoints = tuple(pkg_resources.iter_entry_points('kolekto.datasources', datasource_config.args))
            if not entrypoints:
                raise KolektoRuntimeError('Bad datasource %r' % datasource_config.args)
            datasource_class = entrypoints[0].load()
            datasource_config = datasource_class.config_schema.validate(datasource_config)
            ds = datasource_class(entrypoints[0].name, tree, datasource_config, self._object_class)
            self._datasources.append(ds)

    def search(self, title, **kwargs):
        for datasource in self._datasources:
            for movie in datasource.search(title, **kwargs):
                yield datasource, movie

    def refresh(self, movie):
        for datasource in self._datasources:
            refreshed = datasource.refresh(movie)
            if refreshed:
                movie.update(refreshed)
        return movie

    def attach(self, movie_hash, movie):
        for datasource in self._datasources:
            movie = datasource.attach(movie_hash, movie)
        return movie


import os

from kolekto.db import MoviesMetadata


class Command(object):

    """
    Base class for all commands.

    :cvar help: the help for the command
    """

    help = ''

    def __init__(self, name, profile, aparser_subs):
        self._aparser = aparser_subs.add_parser(name, help=self.help)
        self._profile = profile
        self._aparser.set_defaults(command=self.run, command_name=name)

    @property
    def profile(self):
        return self._profile

    def add_arg(self, *args, **kwargs):
        """ Add an argument to the command argument parser.
        """

        self._aparser.add_argument(*args, **kwargs)

    def prepare(self):
        """ Method to override, executed before to parse arguments from command
            line. This is a good place to call :meth:`add_arg`.
        """
        pass

    def run(self, args, config):
        """ Method to override, executed if command has been selected.

        :param args: parsed arguments
        :param config: parsed configuration
        """
        pass

    def get_metadata_db(self, tree):
        return MoviesMetadata(os.path.join(tree, '.kolekto', 'metadata.db'),
                              object_class=self.profile.object_class)